@@ -1,6 +1,5 @@
 /*  This file is part of the Vc library. {{{
 Copyright © 2009-2015 Matthias Kretz <kretz@kde.org>
-All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
@@ -87,15 +86,13 @@ template <typename T> class Vector<T, VectorAbi::Sse>
         using mask_type = Mask;
         typedef typename Mask::Argument MaskArg;
         typedef typename Mask::Argument MaskArgument;
-#ifdef Vc_PASSING_VECTOR_BY_VALUE_IS_BROKEN
-        typedef const Vector &AsArg;
-#else
         typedef const Vector AsArg;
-#endif
         using abi = VectorAbi::Sse;
         using WriteMaskedVector = Common::WriteMaskedVector<Vector, Mask>;
         template <typename U> using V = Vector<U, abi>;
 
+        using reference = Detail::ElementReference<Vector>;
+
 #include "../common/generalinterface.h"
 
         static Vc_INTRINSIC_L Vector Random() Vc_INTRINSIC_R;
@@ -107,22 +104,25 @@ template <typename T> class Vector<T, VectorAbi::Sse>
         // implict conversion from compatible Vector<U>
         template <typename U>
         Vc_INTRINSIC Vector(
-            Vc_ALIGNED_PARAMETER(V<U>) x,
-            typename std::enable_if<Traits::is_implicit_cast_allowed<U, T>::value,
-                                    void *>::type = nullptr)
+            V<U> x, typename std::enable_if<Traits::is_implicit_cast_allowed<U, T>::value,
+                                            void *>::type = nullptr)
             : d(SSE::convert<U, T>(x.data()))
         {
         }
 
+#if Vc_IS_VERSION_1
         // static_cast from the remaining Vector<U>
         template <typename U>
-        Vc_INTRINSIC explicit Vector(
-            Vc_ALIGNED_PARAMETER(V<U>) x,
-            typename std::enable_if<!Traits::is_implicit_cast_allowed<U, T>::value,
-                                    void *>::type = nullptr)
+        Vc_DEPRECATED("use simd_cast instead of explicit type casting to convert between "
+                      "vector types") Vc_INTRINSIC
+            explicit Vector(
+                V<U> x,
+                typename std::enable_if<!Traits::is_implicit_cast_allowed<U, T>::value,
+                                        void *>::type = nullptr)
             : d(SSE::convert<U, T>(x.data()))
         {
         }
+#endif
 
         ///////////////////////////////////////////////////////////////////////////////////////////
         // broadcast
@@ -158,10 +158,31 @@ template <typename T> class Vector<T, VectorAbi::Sse>
         Vc_INTRINSIC Vector operator++(int) { const Vector r = *this; data() = HT::add(data(), HT::one()); return r; }
         Vc_INTRINSIC Vector operator--(int) { const Vector r = *this; data() = HT::sub(data(), HT::one()); return r; }
 
-        Vc_INTRINSIC decltype(d.ref(0)) operator[](size_t index) { return d.ref(index); }
-        Vc_INTRINSIC_L EntryType operator[](size_t index) const Vc_PURE Vc_INTRINSIC_R;
+    private:
+        friend reference;
+        Vc_INTRINSIC static value_type get(const Vector &o, int i) noexcept
+        {
+            return o.d.m(i);
+        }
+        template <typename U>
+        Vc_INTRINSIC static void set(Vector &o, int i, U &&v) noexcept(
+            noexcept(std::declval<value_type &>() = v))
+        {
+            o.d.set(i, v);
+        }
 
-        Vc_INTRINSIC_L Vector operator[](SSE::int_v perm) const Vc_INTRINSIC_R;
+    public:
+        Vc_ALWAYS_INLINE reference operator[](size_t index) noexcept
+        {
+            static_assert(noexcept(reference{std::declval<Vector &>(), int()}), "");
+            return {*this, int(index)};
+        }
+        Vc_ALWAYS_INLINE value_type operator[](size_t index) const noexcept
+        {
+            return d.m(index);
+        }
+
+        Vc_INTRINSIC_L Vector Vc_VDECL operator[](const SSE::int_v &perm) const Vc_INTRINSIC_R;
 
         Vc_INTRINSIC Vc_PURE Mask operator!() const
         {
@@ -178,16 +199,17 @@ template <typename T> class Vector<T, VectorAbi::Sse>
         Vc_ALWAYS_INLINE_L Vc_PURE_L Vector operator-() const Vc_ALWAYS_INLINE_R Vc_PURE_R;
         Vc_INTRINSIC Vc_PURE Vector operator+() const { return *this; }
 
-        Vc_INTRINSIC_L Vector &operator<<=(AsArg shift)       Vc_INTRINSIC_R;
-        Vc_INTRINSIC_L Vector  operator<< (AsArg shift) const Vc_INTRINSIC_R;
-        Vc_INTRINSIC_L Vector &operator<<=(  int shift)       Vc_INTRINSIC_R;
-        Vc_INTRINSIC_L Vector  operator<< (  int shift) const Vc_INTRINSIC_R;
-        Vc_INTRINSIC_L Vector &operator>>=(AsArg shift)       Vc_INTRINSIC_R;
-        Vc_INTRINSIC_L Vector  operator>> (AsArg shift) const Vc_INTRINSIC_R;
-        Vc_INTRINSIC_L Vector &operator>>=(  int shift)       Vc_INTRINSIC_R;
-        Vc_INTRINSIC_L Vector  operator>> (  int shift) const Vc_INTRINSIC_R;
+        Vc_ALWAYS_INLINE Vector  Vc_VDECL operator<< (AsArg shift) const { return generate([&](int i) { return get(*this, i) << get(shift, i); }); }
+        Vc_ALWAYS_INLINE Vector  Vc_VDECL operator>> (AsArg shift) const { return generate([&](int i) { return get(*this, i) >> get(shift, i); }); }
+        Vc_ALWAYS_INLINE Vector &Vc_VDECL operator<<=(AsArg shift) { return *this = *this << shift; }
+        Vc_ALWAYS_INLINE Vector &Vc_VDECL operator>>=(AsArg shift) { return *this = *this >> shift; }
 
-        Vc_INTRINSIC Vc_PURE Vc_DEPRECATED("use isnegative(x) instead") Mask
+        Vc_INTRINSIC_L Vector &Vc_VDECL operator<<=(  int shift)       Vc_INTRINSIC_R;
+        Vc_INTRINSIC_L Vector  Vc_VDECL operator<< (  int shift) const Vc_INTRINSIC_R;
+        Vc_INTRINSIC_L Vector &Vc_VDECL operator>>=(  int shift)       Vc_INTRINSIC_R;
+        Vc_INTRINSIC_L Vector  Vc_VDECL operator>> (  int shift) const Vc_INTRINSIC_R;
+
+        Vc_DEPRECATED("use isnegative(x) instead") Vc_INTRINSIC Vc_PURE Mask
             isNegative() const
         {
             return Vc::isnegative(*this);
@@ -200,21 +222,19 @@ template <typename T> class Vector<T, VectorAbi::Sse>
         }
 
         template <typename V2>
-        Vc_ALWAYS_INLINE Vc_PURE
-            Vc_DEPRECATED("Use simd_cast instead of Vector::staticCast") V2
-            staticCast() const
+        Vc_DEPRECATED("Use simd_cast instead of Vector::staticCast")
+            Vc_ALWAYS_INLINE Vc_PURE V2 staticCast() const
         {
             return SSE::convert<T, typename V2::EntryType>(data());
         }
         template <typename V2>
-        Vc_ALWAYS_INLINE Vc_PURE
-            Vc_DEPRECATED("use reinterpret_components_cast instead") V2
-            reinterpretCast() const
+        Vc_DEPRECATED("use reinterpret_components_cast instead")
+            Vc_ALWAYS_INLINE Vc_PURE V2 reinterpretCast() const
         {
             return SSE::sse_cast<typename V2::VectorType>(data());
         }
 
-        Vc_INTRINSIC WriteMaskedVector operator()(const Mask &k) { return {this, k}; }
+        Vc_INTRINSIC WriteMaskedVector operator()(const Mask &k) { return {*this, k}; }
 
         Vc_ALWAYS_INLINE Vc_PURE VectorType &data() { return d.v(); }
         Vc_ALWAYS_INLINE Vc_PURE const VectorType &data() const { return d.v(); }
@@ -287,13 +307,13 @@ template <typename T> class Vector<T, VectorAbi::Sse>
 
         template <typename G> static Vc_INTRINSIC_L Vector generate(G gen) Vc_INTRINSIC_R;
 
-        Vc_INTRINSIC Vc_DEPRECATED("use copysign(x, y) instead") Vector
+        Vc_DEPRECATED("use copysign(x, y) instead") Vc_INTRINSIC Vector
             copySign(AsArg reference) const
         {
             return Vc::copysign(*this, reference);
         }
 
-        Vc_INTRINSIC Vc_DEPRECATED("use exponent(x) instead") Vector exponent() const
+        Vc_DEPRECATED("use exponent(x) instead") Vc_INTRINSIC Vector exponent() const
         {
             return Vc::exponent(*this);
         }
